﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using MeshVR;
using static Oni;
using UnityEngine.SceneManagement;
using static OVRProfile;

/*
    Control of water material parammeters.

	Made by DemonLord
	License: Creative Commons with Attribution (CC BY)
*/

namespace DLPlugin {
	public class WaterConfig : MVRScript {

        protected JSONStorableFloat jsScaleX;
        protected JSONStorableFloat jsScaleY;
        protected JSONStorableFloat jsScaleZ;
        protected JSONStorableStringChooser jpopStencilRef;
        protected JSONStorableBool r1Enable;
        protected List<JSONStorableFloat> jFP = new List<JSONStorableFloat>(); //jsonFloatParams
        protected List<JSONStorableColor> jCP = new List<JSONStorableColor>(); //jsonColorParams

        protected JSONStorableBool jEnableRealtimeReflections;
        protected JSONStorableBool jEnableProbeRelfections;

        protected List<MatPackage> origMaterials = new List<MatPackage>();
		protected List<MatPackage> newMaterials = new List<MatPackage>();
		protected List<Renderer> allRenderers = new List<Renderer>();
		protected CustomUnityAssetLoader cUAL;
        protected List<string> layers32 = new List<string>();
        protected bool assetChanged;
        private Transform rescaleObject;
        private Vector3 scale;
        private Vector3 scaleWas;
        protected ReflectionProbe r1;

        public struct MatPackage {
			public int rendererIndex;
			public int materialIndex;
			public Material material;
			public bool isConverted;

			public MatPackage(int rIndex, int mIndex, Material mat, bool converted) {
				this.rendererIndex = rIndex;
				this.materialIndex = mIndex;
				this.material = mat;
				this.isConverted = converted;
			}
		}

		public override void Init() {
			try
            {
                for (int i = 0; i < 32; i++)
                {
                    layers32.Add(i.ToString());
                }
                for (int i = 0; i < containingAtom.childAtomContainer.childCount; i++)
                {
                    if (containingAtom.childAtomContainer.GetChild(i).name == "rescaleObject")
                    {
                        rescaleObject = containingAtom.childAtomContainer.GetChild(i);
                        scale = new Vector3(rescaleObject.transform.localScale.x, rescaleObject.transform.localScale.y, rescaleObject.transform.localScale.z);
                        scaleWas = scale;
                    }
                    break;
                }

                cUAL = containingAtom.childAtomContainer.GetComponentInChildren<CustomUnityAssetLoader>();
                // get the loader component
                SetCUACallback(cUAL);
                //SuperController.LogMessage("allRenderers: " + allRenderers.Count);
			}
			catch (Exception e)
			{
				SuperController.LogError("WaterConfig: " + e);
			}
		}
		
        protected void SetCUACallback(CustomUnityAssetLoader cUAL)
        {
            if (cUAL != null)
            {
                cUAL.RegisterAssetClearedCallback(AssetCleared);
                cUAL.RegisterAssetLoadedCallback(AssetLoaded);
                if (cUAL.isAssetLoaded)
                {
                    AssetLoaded();
                }
            }
        }
		


        protected void BuildUIControls() {
            
            for (int i = 0; i < newMaterials.Count; i++)
            {
                if (newMaterials[i].material.shader.name == "AQUAS-Lite/Frontface")
                {
                    Material mat = newMaterials[i].material;

                    bool val = true;
                    if (mat.GetFloat("_EnableRealtimeReflections") == 0) val = false;
                    jEnableRealtimeReflections = new JSONStorableBool("_EnableRealtimeReflections", val, SetBoolParam);
                    RegisterBool(jEnableRealtimeReflections);
                    CreateToggle(jEnableRealtimeReflections, true);
                    if (mat.GetFloat("_EnableProbeRelfections") == 0) val = false;
                    jEnableProbeRelfections = new JSONStorableBool("_EnableProbeRelfections", val, SetBoolParam);
                    RegisterBool(jEnableProbeRelfections);
                    CreateToggle(jEnableProbeRelfections, true);

                    Color c = mat.GetColor("_MainColor");
                    HSVColor cHSVC = HSVColorPicker.RGBToHSV(c.r, c.g, c.b);
                    jCP.Add(new JSONStorableColor("_MainColor", cHSVC, SetColorParam));
                    c = mat.GetColor("_DeepWaterColor");
                    cHSVC = HSVColorPicker.RGBToHSV(c.r, c.g, c.b);
                    jCP.Add(new JSONStorableColor("_DeepWaterColor", cHSVC, SetColorParam));
                    c = mat.GetColor("_SpecularColor");
                    cHSVC = HSVColorPicker.RGBToHSV(c.r, c.g, c.b);
                    jCP.Add(new JSONStorableColor("_SpecularColor", cHSVC, SetColorParam));
                    c = mat.GetColor("_FoamColor");
                    cHSVC = HSVColorPicker.RGBToHSV(c.r, c.g, c.b);
                    jCP.Add(new JSONStorableColor("_FoamColor", cHSVC, SetColorParam));

                    jpopStencilRef = new JSONStorableStringChooser("StencilRef", new List<string>(layers32), "0", "StencilRef", SetpopStencilRef);
                    RegisterStringChooser(jpopStencilRef);
                    CreateScrollablePopup(jpopStencilRef);
                    jFP.Add(new JSONStorableFloat("scaleX", rescaleObject.transform.localScale.x, SetScaleXYZ, 0f, 15f, true));
                    jFP.Add(new JSONStorableFloat("scaleY", rescaleObject.transform.localScale.y, SetScaleXYZ, 0f, 15f, true));
                    jFP.Add(new JSONStorableFloat("scaleZ", rescaleObject.transform.localScale.z, SetScaleXYZ, 0f, 15f, true));
                    jFP.Add(new JSONStorableFloat("_NormalStrength", mat.GetFloat("_NormalStrength"), SetFloatParam, 0f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_NormalTiling", mat.GetFloat("_NormalTiling"), SetFloatParam, 0.01f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_NormalTiling2", mat.GetFloat("_NormalTiling2"), SetFloatParam, 0.01f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_NormalTiling3", mat.GetFloat("_NormalTiling3"), SetFloatParam, 0.01f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed", mat.GetFloat("_WaveSpeed"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed2", mat.GetFloat("_WaveSpeed2"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed3", mat.GetFloat("_WaveSpeed3"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed4", mat.GetFloat("_WaveSpeed4"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed5", mat.GetFloat("_WaveSpeed5"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_WaveSpeed6", mat.GetFloat("_WaveSpeed6"), SetFloatParam, -3f, 3f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate", mat.GetFloat("_Rotate"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate2", mat.GetFloat("_Rotate2"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate3", mat.GetFloat("_Rotate3"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate4", mat.GetFloat("_Rotate4"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate5", mat.GetFloat("_Rotate5"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Rotate6", mat.GetFloat("_Rotate6"), SetFloatParam, -180f, 180f, true));
                    jFP.Add(new JSONStorableFloat("_Density", mat.GetFloat("_Density"), SetFloatParam, 0f, 1f, true));
                    jFP.Add(new JSONStorableFloat("_Fade", mat.GetFloat("_Fade"), SetFloatParam, 0f, 5f, true));
                    jFP.Add(new JSONStorableFloat("_DepthTransparency", mat.GetFloat("_DepthTransparency"), SetFloatParam, 0f, 200f, true));
                    jFP.Add(new JSONStorableFloat("_TransparencyFade", mat.GetFloat("_TransparencyFade"), SetFloatParam, 0f, 10f, true));
                    jFP.Add(new JSONStorableFloat("_Refraction", mat.GetFloat("_Refraction"), SetFloatParam, 0f, 1f, true));
                    jFP.Add(new JSONStorableFloat("_Specular", mat.GetFloat("_Specular"), SetFloatParam, 0f, 5f, true));
                    jFP.Add(new JSONStorableFloat("_Gloss", mat.GetFloat("_Gloss"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_LightWrapping", mat.GetFloat("_LightWrapping"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_FoamTiling", mat.GetFloat("_FoamTiling"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_FoamVisibility", mat.GetFloat("_FoamVisibility"), SetFloatParam, 0f, 1f, true));
                    jFP.Add(new JSONStorableFloat("_FoamBlend", mat.GetFloat("_FoamBlend"), SetFloatParam, 0f, 10f, true));
                    jFP.Add(new JSONStorableFloat("_FoamContrast", mat.GetFloat("_FoamContrast"), SetFloatParam, 0f, 0.499f, true));
                    jFP.Add(new JSONStorableFloat("_FoamIntensity", mat.GetFloat("_FoamIntensity"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_FoamSpeed", mat.GetFloat("_FoamSpeed"), SetFloatParam, -2f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_RealtimeReflectionIntensity", mat.GetFloat("_RealtimeReflectionIntensity"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_ProbeReflectionIntensity", mat.GetFloat("_ProbeReflectionIntensity"), SetFloatParam, 0f, 2f, true));
                    jFP.Add(new JSONStorableFloat("_Distortion", mat.GetFloat("_Distortion"), SetFloatParam, 0f, 1f, true));
                    jFP.Add(new JSONStorableFloat("_MediumTilingDistance", mat.GetFloat("_MediumTilingDistance"), SetFloatParam, 0f, 2000f, true));
                    jFP.Add(new JSONStorableFloat("_FarTilingDistance", mat.GetFloat("_FarTilingDistance"), SetFloatParam, 0f, 2000f, true));
                    jFP.Add(new JSONStorableFloat("_DistanceFade", mat.GetFloat("_DistanceFade"), SetFloatParam, 0f, 2, true));

                    for (int u = 0; u < jCP.Count; u++)
                    {
                        RegisterColor(jCP[u]);
                        CreateColorPicker(jCP[u], true);
                    }
                    for (int u = 0; u < jFP.Count; u++)
                    {
                        RegisterFloat(jFP[u]);
                        if (jFP[u].name.Contains("scale"))
                        {
                            CreateSlider(jFP[u],true);
                        }
                        else
                        {
                            CreateSlider(jFP[u]);
                        }
                    }
                    // r1EnableToggle
                    r1 = containingAtom.childAtomContainer.GetComponentInChildren<ReflectionProbe>(true);
                    if (r1 != null)
                    {
                        r1Enable = new JSONStorableBool("Reflectionprobe", r1.gameObject.activeSelf, (bool b) => { if (r1 != null) r1.enabled = b; });
                        RegisterBool(r1Enable);
                        CreateToggle(r1Enable, false);
                    }
                }
            }
            
        }
		protected void AssetCleared() {
			// don't need to reset materials because the asset would have been destroyed
			allRenderers.Clear();
            origMaterials.Clear();
			newMaterials.Clear();
        }

        protected void AssetLoaded()
        {
            containingAtom.childAtomContainer.GetComponentsInChildren<Renderer>(false, allRenderers);
            ConvertMaterial();
        }


		protected void ConvertMaterial() {
			if (allRenderers.Count == 0) {
                SuperController.LogMessage("WaterConfig: Make sure you have an waterplane.assetbundle loaded and a prop selected from the Asset dropdown. Then reload this plugin.");
				return;
			}
            // fill newMaterials with copy of the original material or an new material with the Skybox/6 Sided shader
            for (int i = 0; i < allRenderers.Count; i++) {
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
                    origMaterials.Add(new MatPackage(i, j, allRenderers[i].materials[j], false));
                    Material mat = Material.Instantiate(allRenderers[i].materials[j]);
                    newMaterials.Add(new MatPackage(i, j, mat, true));
				}
			}

			if (newMaterials.Count == 0) {
				SuperController.LogError("WaterConfig: No compatible material could be found on this asset. Check message log for skipped materials/shaders.");
				return;
			}

            BuildUIControls();

            // replace renders from original to copy materials
            for (int i = 0; i < allRenderers.Count; i++) {
				Material[] matsToAssign = allRenderers[i].materials;
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
                    Material vamMat = GetVamMaterialByIndex(i, j);
                    //vamMat.CopyPropertiesFromMaterial(GetVamMaterialByIndex(i,j));
                    matsToAssign[j] = vamMat;
                }
				// assign new materials to prop renderer as array
				allRenderers[i].materials = matsToAssign;
            }
            //SuperController.LogMessage("origMaterials: " + origMaterials.Count);
            //SuperController.LogMessage("newMaterials: " + newMaterials.Count);

            JSONNode node;
            if (GetStoredNode("StencilRef", out node))
            {
                //if (node == null) continue;
                jpopStencilRef.valNoCallback = node.Value;
            }
            foreach (JSONStorableColor colorN in jCP)
            {
                RegisterColor(colorN);
                //SuperController.LogMessage("convert 1M: name "+ colorN.name+" val ");
                if (GetStoredNode(colorN.name, out node))
                {
                    //if (node == null) continue;
                    HSVColor cN;
                    cN.H = node["h"].AsFloat;
                    cN.S = node["s"].AsFloat;
                    cN.V = node["v"].AsFloat;
                    colorN.valNoCallback = cN;
                    //SuperController.LogMessage("convert M: name "+ colorN.name+" h "+node["h"]+" s "+node["s"]+" v "+node["v"]);
                }
            }
            foreach (JSONStorableFloat jP in jFP)
            {
                if (GetStoredNode(jP.name, out node))
                {
                    //if (node == null) continue;
                    jP.valNoCallback = float.Parse(node.Value);
                }
            }
            foreach (JSONStorableFloat jP in jFP) SetFloatParam(jP);
            SetBoolParam(jEnableRealtimeReflections);
            SetBoolParam(jEnableProbeRelfections);
            foreach (JSONStorableColor colorN in jCP) SetColorParam(colorN);

            // apply user definable values initially
            SetpopStencilRef(jpopStencilRef.val);
            assetChanged = true;
        }
        public bool GetStoredNode(string storableName, out JSONNode node)
        {
            node = null;
            //SuperController.LogMessage(Type(SuperController.singleton.loadJson));
            // if((SuperController.singleton.loadJson?["atoms"] ?? null) == null) SuperController.LogMessage("11");
            if (assetChanged || (SuperController.singleton.loadJson?["atoms"] ?? null) == null) return false;
            foreach (JSONNode a in SuperController.singleton.loadJson["atoms"].AsArray)
            {
                if (a["id"].Value == containingAtom.name)
                {
                    foreach (JSONNode n in a["storables"].AsArray)
                    {
                        if (n["id"].Value.Contains("DLPlugin.WaterConfig"))
                        {
                            //SuperController.LogMessage(storableName + " val "+n[storableName].Value);
                            if (n[storableName] != null)
                            {
                                node = n[storableName];
                                return true;
                            }
                        }
                    }
                }
            }
            return false;
        }
        protected void SetScaleXYZ(JSONStorableFloat jf)
        {
            if (jf.name == "scaleX") scale.x = jf.val;
            if (jf.name == "scaleY") scale.y = jf.val;
            if (jf.name == "scaleZ") scale.z = jf.val;
            rescaleObject.transform.localScale = scale;
            //SuperController.LogMessage("scale "+ scale.x+ " "+ scale.y+ " "+ scale.z);
        }

        protected void SetpopStencilRef(string v)
        {
            if (allRenderers.Count > 0)
            {
                foreach (MatPackage matPackage in newMaterials)
                {
                    if (!matPackage.isConverted) continue;
                    if (matPackage.material.HasProperty("_StencilRef"))
                        matPackage.material.SetInt("_StencilRef", Int32.Parse(v));
                }
            }
            
        }
        protected Material GetVamMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in newMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		void OnDestroy()
        {
            rescaleObject.transform.localScale = scaleWas;
            if (r1 != null) r1.enabled = true;
            RestoreOriginalMaterials();

            if (cUAL != null)
            {
                cUAL.DeregisterAssetClearedCallback(AssetCleared);
                cUAL.DeregisterAssetLoadedCallback(AssetLoaded);
            }
        }

		protected Material GetOrigMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in origMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		protected void RestoreOriginalMaterials() {
			for (int i = 0; i < allRenderers.Count; i++) {
				if (allRenderers[i] != null) {   // renderer may have been destroyed already by atom remove
					List<Material> matsToAssign = new List<Material>();
					for (int j = 0; j < allRenderers[i].materials.Length; j++) {
						matsToAssign.Add(GetOrigMaterialByIndex(i, j));
					}
					allRenderers[i].materials = matsToAssign.ToArray();
				}
			}
        }
        protected void SetBoolParam(JSONStorableBool jf)
        {
            float val = 0;
            foreach (MatPackage matN in newMaterials)
            {
                if (matN.material.HasProperty(jf.name)) {
                    if (jf.val) val = 1f;
                    matN.material.SetFloat(jf.name, val);
                }
            }
        }

        protected void SetFloatParam(JSONStorableFloat jf) {
            if (jf.name.Contains("scale"))
            {
                SetScaleXYZ(jf);
                return;
            }
			foreach (MatPackage matN in newMaterials) {
                if (matN.material.HasProperty(jf.name)) matN.material.SetFloat(jf.name, jf.val);
			}
		}

		protected void SetColorParam(JSONStorableColor jcolor) {
			foreach (MatPackage matN in newMaterials)
            {
                if (matN.material.HasProperty(jcolor.name))
                {
                    Color c = jcolor.colorPicker?.currentColor ?? HSVColorPicker.HSVToRGB(jcolor.val);
                    c.a = matN.material.GetColor(jcolor.name).a;
                    matN.material.SetColor(jcolor.name, c);
                    break;
                }
            }
		}


	}
}